#pragma once
#include <iostream>
#include <cstring>
#include <cassert>
#include <atl/object.hpp>

namespace cpptesk
{
namespace ts
{

enum scenario_function_mode { INIT_ITERATION, NEXT_ITERATION, INVOKE_ACTION };

/** \brief Base class for iteration contextes.
 * All classes, implementing specific contextes, must derive from this class.
 */
struct ContextBase
{
    /// Internal variable. Do not modify it.
    int __line;
    /// Internal variable. Do not modify it.
    int __action_line;
    /// Internal variable. Do not modify it.
    bool __direct_jump;
    /// Internal variable. Do not modify it.
    int __dummy_loop;
    /// Internal variable. Do not modify it.
    scenario_function_mode __mode;

    ContextBase() : __line(0), __action_line(-1), __direct_jump(false), __mode(INIT_ITERATION) {}
    /// Internal method. Do not call.
    void enter_action(int line) { assert(line > 0); __action_line = line; }
    /// Internal method. Do not call.
    void leave_action() { assert(__action_line > 0); __line = __action_line; }
};

struct IntCtx : public ContextBase
{
    int i;
    IntCtx() : i(0) {}
};
inline bool operator == (const IntCtx& left, const IntCtx& right)
{
    return left.i == right.i;
}
inline std::ostream& operator<< (std::ostream& out, const IntCtx& ref)
{
    return out << ref.i;
}

struct Int2Ctx : public ContextBase
{
    int i;
    int j;
    Int2Ctx() : i(0),j(0) {}
};
inline bool operator == (const Int2Ctx& left, const Int2Ctx& right)
{
    return left.i == right.i && left.j == right.j;
}
inline std::ostream& operator<< (std::ostream& out, const Int2Ctx& ref)
{
    return out << ref.i << ',' << ref.j;
}

struct Int3Ctx : public ContextBase
{
    int i;
    int j;
    int k;
    Int3Ctx() : i(0),j(0),k(0) {}
};
inline bool operator == (const Int3Ctx& left, const Int3Ctx& right)
{
    return left.i == right.i && left.j == right.j && left.k == right.k;
}
inline std::ostream& operator<< (std::ostream& out, const Int3Ctx& ref)
{
    return out << ref.i << ',' << ref.j << ',' << ref.k;
}

struct IntAbcCtx : public ContextBase
{
    int a;
    int b;
    int c;
    int d;
    int e;
    int f;
    int g;
    int h;
    int i;
    int j;
    int k;
    int l;
    int m;
    int n;
    int o;
    int p;
    int q;
    int r;
    int s;
    int t;
    int u;
    int v;
    int w;
    int x;
    int y;
    int z;
    IntAbcCtx()
    {
        memset(this, 0, sizeof(IntAbcCtx));
    }
};
inline bool operator == (const IntAbcCtx& left, const IntAbcCtx& right)
{
    return !memcmp(&left, &right, sizeof(IntAbcCtx));
}
inline std::ostream& operator<< (std::ostream& out, const IntAbcCtx& ref)
{
    return out << ref.a << ',' << ref.b << ',' << ref.c
           << ref.d << ',' << ref.e << ',' << ref.f
           << ref.g << ',' << ref.h << ',' << ref.i
           << ref.j << ',' << ref.k << ',' << ref.l
           << ref.m << ',' << ref.n << ',' << ref.o
           << ref.p << ',' << ref.q << ',' << ref.r
           << ref.s << ',' << ref.t << ',' << ref.u
           << ref.v << ',' << ref.w << ',' << ref.x
           << ref.y << ',' << ref.z;
}

} // namespace cpptesk::ts
} // namespace cpptesk

namespace cpptesk
{
namespace atl
{
template<>
struct _Storage<cpptesk::ts::IntCtx> : public _InPlace<cpptesk::ts::IntCtx> {};
template<>
struct _Storage<cpptesk::ts::Int2Ctx> : public _InPlace<cpptesk::ts::Int2Ctx> {};
template<>
struct _Storage<cpptesk::ts::Int3Ctx> : public _InPlace<cpptesk::ts::Int3Ctx> {};
template<>
struct _Storage<cpptesk::ts::IntAbcCtx> : public _InPlace<cpptesk::ts::IntAbcCtx> {};
} // namespace cpptesk::atl
} // namespace cpptesk

//
// Scenario function or method as a co-function or co-method
//

#define CPPTESK_ITERATION__START_LABEL -1
#define CPPTESK_ITERATION__END_LABEL -2

#if !defined(CPPTESK_ITERATION__LOCAL_VARS_IN_ACTION) || defined(CPPTESK_ITERATION__OLD_ACTION)
/// @brief Macro that returns the specified value from a co-routine.
///
/// @param x value to return
/// @return the value of @a param
///
/// @note Co-routine macros must be at most one on a line of code.
#define CPPTESK_ITERATION_YIELD(x) do { ctx.__line = __LINE__; return (x); case __LINE__:; } while(0)

/// @brief Macro that initiates iteration in a co-routine.
///
/// @note Co-routine macros must be at most one on a line of code.
#define CPPTESK_ITERATION_BEGIN \
	if (ctx.__mode == cpptesk::ts::INIT_ITERATION) { ctx.__line = CPPTESK_ITERATION__START_LABEL; }\
	switch(ctx.__line) { \
	case CPPTESK_ITERATION__START_LABEL:\

/// @brief Macro that initiates iteration in a co-routine if condition holds.
///
/// @note Co-routine macros must be at most one on a line of code.
#define CPPTESK_ITERATION_BEGIN_IF(condition) \
	if (ctx.__mode == cpptesk::ts::INIT_ITERATION) { ctx.__line = CPPTESK_ITERATION__START_LABEL; if (!(condition)) return false; }\
	switch(ctx.__line) { \
	case CPPTESK_ITERATION__START_LABEL:\

/// @brief Macro that depicts action block in a co-routine.
///
/// @note Co-routine macros must be at most one on a line of code.
#define CPPTESK_ITERATION_ACTION \
	if (ctx.__mode == cpptesk::ts::INIT_ITERATION) { ctx.__line = 0; return true; case 0:; } \
	if (ctx.__mode == cpptesk::ts::NEXT_ITERATION) { CPPTESK_ITERATION_YIELD(true); } \
	if (ctx.__mode == cpptesk::ts::INVOKE_ACTION) \

/// @brief Macro that ends iteration in a co-routine.
///
/// @note Co-routine macros must be at most one on a line of code.
#define CPPTESK_ITERATION_END default: break;}	return (ctx.__mode == cpptesk::ts::NEXT_ITERATION)? false : true;

#else // CPPTESK_ITERATION__LOCAL_VARS_IN_ACTION

//    IACTION       

#define CPPTESK_ITERATION_YIELD(x) do { ctx.__line = ctx.__action_line; return (x); } while(0)

#define CPPTESK_ITERATION_BEGIN \
	if (ctx.__mode == cpptesk::ts::INIT_ITERATION) { ctx.__direct_jump = false; ctx.__line = CPPTESK_ITERATION__START_LABEL; }\
	else { ctx.__direct_jump = true; } \
	switch(ctx.__line) { \
	case CPPTESK_ITERATION__START_LABEL:\
	for (ctx.__dummy_loop = 0; ctx.__dummy_loop < 1; ctx.__dummy_loop ++) { \

#define CPPTESK_ITERATION_BEGIN_IF(condition) \
	if (ctx.__mode == cpptesk::ts::INIT_ITERATION) { ctx.__line = CPPTESK_ITERATION__START_LABEL; if (!(condition)) return false; }\
	else { ctx.__direct_jump = true; } \
	switch(ctx.__line) { \
	case CPPTESK_ITERATION__START_LABEL:\
	for (ctx.__dummy_loop = 0; ctx.__dummy_loop < 1; ctx.__dummy_loop ++) { \

#define CPPTESK_ITERATION_ACTION \
    case __LINE__: \
    if (ctx.__mode == cpptesk::ts::NEXT_ITERATION && ctx.__direct_jump ) \
        { ctx.__direct_jump = false; continue; } \
     \
	else if (ctx.__mode == cpptesk::ts::INIT_ITERATION) { ctx.__action_line = __LINE__; CPPTESK_ITERATION_YIELD(true); } \
	else if (ctx.__mode == cpptesk::ts::NEXT_ITERATION) { ctx.__action_line = __LINE__; CPPTESK_ITERATION_YIELD(true); } \
	else \

#define CPPTESK_ITERATION_END \
	} \
    default: break;}\
	return (ctx.__mode == cpptesk::ts::NEXT_ITERATION)? false : true;

// case CPPTESK_ITERATION__END_LABEL: default: ctx.__line = CPPTESK_ITERATION__END_LABEL
#endif

/// @brief Macro that returns the reference to the specified iteration variable in a co-routine.
///
/// @note Co-routine macros must be at most one on a line of code.
#define CPPTESK_ITERATION_VARIABLE(name) ctx.name

#if (defined(CPPTESK_SHORT_NAMES) || defined(CPPTESK_ITERATION_SHORT_NAMES)) \
    && !defined(CPPTESK_ITERATION_NO_SHORT_NAMES)
#define ITERATION_YIELD CPPTESK_ITERATION_YIELD
#define ITERATION_VAR CPPTESK_ITERATION_VARIABLE
#define ITERATION_BEGIN CPPTESK_ITERATION_BEGIN
#define ITERATION_BEGIN_IF CPPTESK_ITERATION_BEGIN_IF
#define ITERATION_ACTION CPPTESK_ITERATION_ACTION
#define ITERATION_END CPPTESK_ITERATION_END
#endif

#if (defined(CPPTESK_SHORT_SHORT_NAMES) || defined(CPPTESK_ITERATION_SHORT_SHORT_NAMES)) \
    && !defined(CPPTESK_ITERATION_NO_SHORT_NAMES)
#define YIELD CPPTESK_ITERATION_YIELD
#define IVAR CPPTESK_ITERATION_VARIABLE
#define IBEGIN CPPTESK_ITERATION_BEGIN
#define IBEGIN_IF CPPTESK_ITERATION_BEGIN_IF
#define IACTION CPPTESK_ITERATION_ACTION
#define IEND CPPTESK_ITERATION_END
#endif
