/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Reaction arbiters
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/assertion.hpp>
#include <hw/message.hpp>
#include <hw/process.hpp>
#include <hw/unused.hpp>

#include <algorithm>
#include <map>
#include <vector>

namespace cpptesk {
namespace hw {

////////////////////////////////////////////////////////////////////////////////
// Default Reaction Matcher
////////////////////////////////////////////////////////////////////////////////

class ReactionMatcher
{
public:
    ReactionMatcher() {}
    virtual ~ReactionMatcher() {}

    virtual void appendRequest(const Process &process, const Message &message)
    {
        CPPTESK_UNUSED_ARG(process);
	CPPTESK_UNUSED_ARG(message);
    }

    virtual bool chooseRequest(const Process &process, const Message &message) const
    {
        CPPTESK_UNUSED_ARG(process);
	CPPTESK_UNUSED_ARG(message);

        return true;
    }

    virtual void removeRequest(const Process &process, const Message &message)
    {
        CPPTESK_UNUSED_ARG(process);
	CPPTESK_UNUSED_ARG(message);
    }
};

////////////////////////////////////////////////////////////////////////////////
// FIFO Reaction Matcher
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_FIFO_ARBITER Fifo

#ifdef HW_SHORT_NAMES
    #define FIFO_ARBITER CPPTESK_FIFO_ARBITER
#endif // HW_SHORT_NAMES

class FifoReactionMatcher: public ReactionMatcher
{
public:
    FifoReactionMatcher() {}
    virtual ~FifoReactionMatcher() {}

    virtual void appendRequest(const Process &process, const Message &message)
    {
        CPPTESK_UNUSED_ARG(message);
        requests.push_back((Process*)&process);
    }

    virtual bool chooseRequest(const Process &process, const Message &message) const
    {
        CPPTESK_UNUSED_ARG(message);
        CPPTESK_ASSERTION(!requests.empty(), "There are no requests to choose");

        return requests[0] == &process;
    }

    virtual void removeRequest(const Process &process, const Message &message)
    {
        CPPTESK_UNUSED_ARG(message);
        CPPTESK_ASSERTION(!requests.empty(), "There are no requests to remove");

        for(std::vector<Process*>::iterator i = requests.begin(); i != requests.end(); i++)
            { if(*i == &process) { requests.erase(i); break; } }
    }

protected:
    /// Queue of reaction requests.
    std::vector<Process*> requests;
};

////////////////////////////////////////////////////////////////////////////////
// Priority Reaction Matcher
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_PRIORITY_ARBITER Priority

#ifdef HW_SHORT_NAMES
    #define PRIORITY_ARBITER CPPTESK_PRIORITY_ARBITER
#endif // HW_SHORT_NAMES

class PriorityReactionMatcher: public ReactionMatcher
{
public:
    PriorityReactionMatcher() {}
    virtual ~PriorityReactionMatcher() {}

    virtual void appendRequest(const Process &process, const Message &message)
    {
        CPPTESK_UNUSED_ARG(message);

        requests.push_back((Process*)&process);
        sort(requests.begin(), requests.end(), greaterProcessPriority);
    }

    virtual bool chooseRequest(const Process &process, const Message &message) const
    {
        CPPTESK_UNUSED_ARG(message);
        CPPTESK_ASSERTION(!requests.empty(), "There are no requests to choose");

        return requests[0] == &process;
    }

    virtual void removeRequest(const Process &process, const Message &message)
    {
        CPPTESK_UNUSED_ARG(message);
        CPPTESK_ASSERTION(!requests.empty(), "There are no requests to remove");

        for(std::vector<Process*>::iterator i = requests.begin(); i != requests.end(); i++)
            { if(*i == &process) { requests.erase(i); break; } }
    }

protected:
    /// Queue of reaction requests.
    std::vector<Process*> requests;
};

////////////////////////////////////////////////////////////////////////////////
// Interface Ordering Reaction Matcher
////////////////////////////////////////////////////////////////////////////////

#define CPPTESK_INPUT_FIFO_ARBITER InterfaceOrdering

#ifdef HW_SHORT_NAMES
    #define INPUT_FIFO_ARBITER CPPTESK_INPUT_FIFO_ARBITER
#endif // HW_SHORT_NAMES

class InterfaceOrderingReactionMatcher: public ReactionMatcher
{
public:
    InterfaceOrderingReactionMatcher() {}
    virtual ~InterfaceOrderingReactionMatcher() {}

    virtual void appendRequest(const Process &process, const Message &message)
    {
        CPPTESK_UNUSED_ARG(message);

        std::vector<Process*> &processes = requests[process.input_iface];
        processes.push_back((Process*)&process);
    }

    virtual bool chooseRequest(const Process &process, const Message &message) const
    {
        CPPTESK_UNUSED_ARG(message);

        const std::vector<Process*> processes = requests.find(process.input_iface)->second;
        CPPTESK_ASSERTION(!processes.empty(), "There are no requests to choose");
        
        return processes[0] == &process;
    }

    virtual void removeRequest(const Process &process, const Message &message)
    {
        CPPTESK_UNUSED_ARG(message);

        std::vector<Process*> &processes = requests[process.input_iface];
        CPPTESK_ASSERTION(!processes.empty(), "There are no requests to remove");

        for(std::vector<Process*>::iterator i = processes.begin(); i != processes.end(); i++)
            { if(*i == &process) { processes.erase(i); break; } }
    }

protected:
    typedef std::vector<Process*> request_vector_t;
    std::map<Interface*, request_vector_t> requests;
};

}} // namespace cpptesk::hw

