/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Macroses to define a mediator for a hardware model
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/errors.hpp>
#include <hw/model.hpp>

#include <sstream>

////////////////////////////////////////////////////////////////////////////////
// Declaration of Mediator Classes
////////////////////////////////////////////////////////////////////////////////

/** @def CPPTESK_MEDIA
 * Declares a mediator (adapter) of a hardware model.
 */
#define CPPTESK_MEDIA(UserMedia, UserModel) \
    struct UserMedia: public UserModel

/** @def CPPTESK_SIGNAL_MEDIA
 * Declares a signal mediator (adapter) of a hardware model.
 */
#define CPPTESK_SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs) \
    struct UserMedia: public UserModel, public ::cpptesk::hw::IOSignals<Inputs, Outputs>

/** @def CPPTESK_ADAPTER
 * Alias of CPPTESK_MEDIA
 */
#define CPPTESK_ADAPTER(UserMedia, UserModel) CPPTESK_MEDIA(UserMedia, UserModel)

/** @def CPPTESK_SIGNAL_ADAPTER
 * Alias of CPPTESK_SIGNAL_MEDIA
 */
#define CPPTESK_SIGNAL_ADAPTER(UserMedia, UserModel, Inputs, Outputs) CPPTESK_SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs)

#ifdef HW_SHORT_NAMES
    #define MEDIA(UserMedia, UserModel)                           CPPTESK_MEDIA(UserMedia, UserModel)
    #define ADAPTER(UserMedia, UserModel)                         CPPTESK_ADAPTER(UserMedia, UserModel)
    #define SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs)   CPPTESK_SIGNAL_MEDIA(UserMedia, UserModel, Inputs, Outputs)
    #define SIGNAL_ADAPTER(UserMedia, UserModel, Inputs, Outputs) CPPTESK_SIGNAL_ADAPTER(UserMedia, UserModel, Inputs, Outputs)
#endif // HW_SHORT_NAMES

#define CPPTESK_DEFINE_BASIC_OUTPUT_LISTENER(listener_name, output_iface, condition) \
    CPPTESK_DECLARE_PROCESS(listener_name) { \
        CPPTESK_UNUSED_ARG(CPPTESK_PARAMETER_MESSAGE); \
	CPPTESK_UNUSED_ARG(CPPTESK_PARAMETER_PROCESS); \
        CPPTESK_START_PROCESS(); \
        while(true) { \
            CPPTESK_WAIT(condition); \
            CPPTESK_CYCLE(); \
            if(!iface.isReactionTaken()) { \
                std::stringstream buffer; \
                buffer << CPPTESK_UNEXPECTED_REACTION_STRING << ": " << iface.getName(); \
                registerReaction(::cpptesk::hw::UNEXPECTED_REACTION, iface, buffer.str()); \
            } \
        } \
        CPPTESK_STOP_PROCESS(); \
    }

#define CPPTESK_CALL_OUTPUT_LISTENER(full_listener_name, output_iface) \
    CPPTESK_CALL_LOOP_EX(full_listener_name,  output_iface, ::cpptesk::hw::HIGHEST_PRIORITY)

#ifdef HW_SHORT_NAMES
    #define DEFINE_BASIC_OUTPUT_LISTENER(listener_name, output_iface, condition) CPPTESK_DEFINE_BASIC_OUTPUT_LISTENER(listener_name, output_iface, condition)
    #define CALL_OUTPUT_LISTENER(full_listener_name, output_iface)               CPPTESK_CALL_OUTPUT_LISTENER(full_listener_name, output_iface)
#endif // HW_SHORT_NAMES

/** @def CPPTESK_CAPTURE_IFACE
 * Captures an interface.
 */
#define CPPTESK_CAPTURE_IFACE() \
   do { \
        CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_INFO, "[CAPTURE] (iface=" << CPPTESK_GET_IFACE().getName() << ")"); \
        CPPTESK_GET_IFACE().capture(); \
    } while(false)

/** @def CPPTESK_RELEASE_IFACE
 * Releases an interface.
 */
#define CPPTESK_RELEASE_IFACE() \
     do { \
        CPPTESK_DEBUG_PRINT(::cpptesk::hw::DEBUG_INFO, "[RELEASE] (iface=" << CPPTESK_GET_IFACE().getName() << ")"); \
        CPPTESK_GET_IFACE().release(); \
     } while(false)

#ifdef HW_SHORT_NAMES
    #define CAPTURE_IFACE() CPPTESK_CAPTURE_IFACE()
    #define RELEASE_IFACE() CPPTESK_RELEASE_IFACE()
#endif // HW_SHORT_NAMES

namespace cpptesk {
namespace hw {

}} // namespace cpptesk::hw

