/*
 * Copyright 2010 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Pseudo-random number generator
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <stdint.h>
#include <iostream>

#define CPPTESK_RANDOM(type) \
	(::cpptesk::hw::Random::create().rand<type>())

#define CPPTESK_RANDOM_WIDTH(type, width) \
	(::cpptesk::hw::Random::create().rand<type>(width))

#define CPPTESK_RANDOM_FIELD(type, lo, hi) \
	(::cpptesk::hw::Random::create().rand<type>((unsigned)(lo), (unsigned)(hi)))

#define CPPTESK_RANDOM_RANGE(type, min, max) \
	(::cpptesk::hw::Random::create().rand_range<type>((type)(min), (type)(max)))

#ifdef HW_SHORT_NAMES
	#define RANDOM(type)                 CPPTESK_RANDOM(type)
	#define RANDOM_WIDTH(type, width)    CPPTESK_RANDOM_WIDTH(type, width)
	#define RANDOM_FIELD(type, lo, hi)   CPPTESK_RANDOM_FIELD(type, lo, hi)
	#define RANDOM_RANGE(type, min, max) CPPTESK_RANDOM_RANGE(type, min, max)
#endif

namespace cpptesk {
namespace hw {

class Random
{
public:
    typedef unsigned default_type_t;
    static Random& create();

    template<typename T> static inline T abs(T val)
    {
	    return val & ~((T)1 << (8*sizeof(T) - 1));
    }

    template<typename T> static inline T mask(unsigned width)
    {
        return ((T)1 << width) - 1;
    }

    template<typename T> static inline T mask(unsigned lo, unsigned hi)
    {
        return mask<T>((hi - lo) + 1) << lo;
    }

    template<typename T> inline T rand()
    {
        // Lower bits are more deterministic in linear random generators
        uint64_t res  = sizeof(T) < sizeof(uint32_t) ? next() >> 8 : next();
    
        res &= mask<uint64_t>(8*sizeof(T));

        for(unsigned i = 1; i <= sizeof(T)/4; i++)
            { res <<= 32; res |= next(); }

        return (T)res;
    }

    template<typename T> inline T rand_range(T min, T max)
    {
        T base = abs<T>(rand<T>());

        return min + base % ((max - min) + 1);
    }

    template<typename T> inline T rand_field(unsigned width)
    {
        const unsigned T_width = 8*sizeof(T);

        // Again, lower bits are more deterministic in linear random generators
        return (width <= T_width - 8 ? rand<T>() >> 8 : rand<T>()) & mask<T>(width);
    }

    template<typename T> inline T rand_field(unsigned lo, unsigned hi)
    {
        const unsigned T_width = 8*sizeof(T);

        // Again, lower bits are more deterministic in linear random generators
        return (hi < T_width - 8 ? rand<T>() >> 8 : rand<T>()) & mask<T>(lo, hi);
    }

    template<typename T> inline T rand(unsigned width)           { return rand_field<T>(width);  }
    template<typename T> inline T rand(unsigned lo, unsigned hi) { return rand_field<T>(lo, hi); }

protected:
    Random() {}
    virtual ~Random() {}

    virtual uint32_t next()
    {
        return _value = 1664525 * _value + 1013904223;
    }

    virtual void set_seed(uint32_t seed)
    {
        _value = seed;
    }

private:
    static uint32_t _value;
};

struct Randomizable: public Random
{
    virtual ~Randomizable() {}
    virtual void randomize() = 0;
};

}} // namespace cpptesk::hw

